unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  GLScene, GLObjects, GLCadencer, GLMisc, GLVectorFileObjects, ExtCtrls,
  StdCtrls, GLWin32Viewer, GLTexture, VectorGeometry

  , NewtonImport // Use Newton.

  ;

type
  TForm1 = class(TForm)
    Viewer: TGLSceneViewer;
    GLScene: TGLScene;
    GLCadencer: TGLCadencer;
    GLCamera1: TGLCamera;
    GLLightSource1: TGLLightSource;
    DCCameraTarget: TGLDummyCube;
    Timer1: TTimer;
    GLMaterialLibrary: TGLMaterialLibrary;
    GLLightSource2: TGLLightSource;
    GLFreeForm1: TGLFreeForm;
    Panel1: TPanel;
    DCODE1: TGLDummyCube;
    Button1: TButton;
    Label1: TLabel;
    procedure GLCadencerProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure Timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ViewerMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure ViewerMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);

      // Start button.
    procedure Button1Click(Sender: TObject);

    procedure CreateNewton1;
    procedure DestroyNewton1;
    procedure UpdateNewton1;
  private
    mdx, mdy : Integer;
  public
  end;

  procedure ForceAndTorqueCallback(const Body : PNewtonBody); cdecl;

var
  Form1: TForm1;

    // Newton Physics.
  NewtonWorld   : PNewtonWorld;
  PhysTime      : Double;
  BodyList      : array of PNewtonBody;

implementation

{$R *.DFM}

uses GLFileLOD, ULODkaUtilsA1;        // 1) Use LOD-file.

procedure TForm1.FormCreate(Sender: TObject);
var
  i, bd1, gm1, k1  : Integer;
  ff1              : TGLFreeForm;
  Mesh             : TMeshObject;
  NBoxGeomCol      : PNewtonCollision;
  Inertia          : TAffineVector;
  isWorld          : Boolean;
  Matrix2          : TMatrix;
  MassDest1        : Single;
begin
    // 2) Enable physic loading.
  EnablePhysicsLoading := True;
    // 3) Load LOD-file.
  GLFreeForm1.LoadFromFile('ode_world_01.lod');

    // 4) Read 3 physics array ( LODBodys, LODGeoms, LODJoints ) and build
    //    all Newton objects.

  for i := 0 to GLFreeForm1.MeshObjects.Count -1 do
    GLFreeForm1.MeshObjects[i].Visible := True;
  GLFreeForm1.Visible := False;

  Label1.Caption := Format('Bodys:%d Geoms:%d Joints:%d', [
    Length(LODBodys),
    Length(LODGeoms),
    Length(LODJoints)]);


  CreateNewton1;
  for bd1 := 0 to Length(LODBodys) -1 do begin
    isWorld := CompareText(LODBodys[bd1].Name, 'World') = 0;

      // Find and create geoms ( collision ).
    for gm1 := 0 to Length(LODGeoms) -1 do begin
      if LODGeoms[gm1].BodyIndex <> bd1 then continue;
      if LODGeoms[gm1].Type1 <> bdtBox then continue;

        // Graphics.
      FF1 := nil;
      if     (Length(LODGeoms[gm1].MeshIndex) >  0)
         and (LODGeoms[gm1].MeshIndex[0] >= 0)
         and (LODGeoms[gm1].MeshIndex[0] <= GLFreeForm1.MeshObjects.Count -1)
      then begin
        FF1 := TGLFreeForm(DCODE1.AddNewChild(TGLFreeForm));
        with FF1 do begin
          FF1.MaterialLibrary := GLFreeForm1.MaterialLibrary;
          Mesh := TMeshObject.CreateOwned(FF1.MeshObjects);
          Mesh.Assign(GLFreeForm1.MeshObjects[ LODGeoms[gm1].MeshIndex[0] ]);
        end;
      end;

      NBoxGeomCol := NewtonCreateBox(NewtonWorld,
        LODGeoms[gm1].Scale[0],
        LODGeoms[gm1].Scale[1],
        LODGeoms[gm1].Scale[2], nil);

        // Create the rigid body
      k1 := Length(BodyList);
      SetLength(BodyList, k1 +1);
      BodyList[k1] := NewtonCreateBody(NewtonWorld, NBoxGeomCol);
        // Remove the collider, we don't need it anymore
      NewtonReleaseCollision(NewtonWorld, NBoxGeomCol);

      if isWorld then MassDest1 := 0
                 else MassDest1 := LODBodys[bd1].MassDensity;
      Inertia[0] := MassDest1
        *(LODBodys[bd1].Scale[1] *LODBodys[bd1].Scale[1]
        + LODBodys[bd1].Scale[2] *LODBodys[bd1].Scale[2]) / 12;
      Inertia[1] := MassDest1
        *(LODBodys[bd1].Scale[0] *LODBodys[bd1].Scale[0]
        + LODBodys[bd1].Scale[2] *LODBodys[bd1].Scale[2]) / 12;
      Inertia[2] := MassDest1
        *(LODBodys[bd1].Scale[0] *LODBodys[bd1].Scale[0]
        + LODBodys[bd1].Scale[1] *LODBodys[bd1].Scale[1]) / 12;
      NewtonBodySetMassMatrix(BodyList[k1], MassDest1,
        Inertia[0], Inertia[1], Inertia[2]);

        // Now set the position of the body's matrix
      Matrix2 := QuaternionToMatrix(LODBodys[bd1].Rotation);
      Matrix2[3, 0] := LODBodys[bd1].Position[0];
      Matrix2[3, 1] := LODBodys[bd1].Position[1];
      Matrix2[3, 2] := LODBodys[bd1].Position[2];
      NewtonBodySetMatrix(BodyList[k1], @Matrix2[0, 0]);

        // Finally set the callback in which the forces on this body will be
        //  applied
      NewtonBodySetForceAndTorqueCallBack(BodyList[k1], ForceAndTorqueCallBack);

      NewtonBodySetUserData(BodyList[k1], FF1);
      break;
    end;
  end;

  UpdateNewton1;
end;


procedure TForm1.FormDestroy(Sender: TObject);
begin
  DestroyNewton1;
end;


procedure TForm1.GLCadencerProgress(Sender: TObject; const deltaTime,
  newTime: Double);
begin
  while physTime < newtime do
    UpdateNewton1;
  Viewer.Invalidate;
end;


  // Start button.
procedure TForm1.Button1Click(Sender: TObject);
begin
  Button1.Enabled    := False;
  physTime           := GLCadencer.CurrentTime;
  GLCadencer.Enabled := True;
end;



  // ------------------------- Newton service ---------------------------------
procedure TForm1.UpdateNewton1;
const
  PhysStep = 1 / 50;
var
  b1      : Integer;
  FF1     : TGLFreeForm;
  Matrix1 : TMatrix;
begin
  physTime := physTime +PhysStep/4;

  NewtonUpdate(NewtonWorld, PhysStep);

  for b1 := 0 to Length(BodyList) -1 do
    if BodyList[b1] <> nil then begin
      FF1 := NewtonBodyGetUserData(BodyList[b1]);
      if FF1 <> nil then begin
        NewtonBodyGetMatrix(BodyList[b1], @Matrix1[0, 0]);
        FF1.Matrix := Matrix1;
      end;
    end;
end;


procedure ForceAndTorqueCallback(const Body : PNewtonBody); cdecl;
var
  Mass           : Single;
  Inertia, Force : TAffineVector;
begin
  NewtonBodyGetMassMatrix(Body, @Mass, @Inertia[0], @Inertia[1], @Inertia[2]);
  Force := AffineVectorMake(0, 0, -9.8 *Mass);
  NewtonBodyAddForce(Body, @Force[0]);
end;


procedure TForm1.CreateNewton1;
begin
  NewtonWorld := NewtonCreate(nil, nil);
end;


procedure TForm1.DestroyNewton1;
begin
  NewtonDestroy(NewtonWorld);
end;






  // ----------------------------- service ------------------------------------

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  Caption := Format('%.1f FPS', [Viewer.FramesPerSecond]);
  Viewer.ResetPerformanceMonitor;
end;


procedure TForm1.ViewerMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if Shift = [ssLeft] then GLCamera1.MoveAroundTarget(mdy -y, mdx -x);
  mdx := x;
  mdy := y;
end;


procedure TForm1.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  if Viewer.Focused then
    GLCamera1.AdjustDistanceToTarget(Power(1.03, WheelDelta/120));
end;


procedure TForm1.FormResize(Sender: TObject);
begin
  GLCamera1.FocalLength := Width /10;
end;


procedure TForm1.ViewerMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Viewer.SetFocus;
end;

end.
