unit GLFileLOD;

{: GLFileLOD<p>
   LODka 3d models editor files

     History:
   date  7 jun 2006 - Add physics support.
   date  1 may 2006 - Add skeleton support.
   date 23 feb 2006 - Creation.
}

interface

uses
  Classes, SysUtils, GLVectorFileObjects, GLTexture, ApplicationFileIO,
  VectorGeometry
  , ULODkaUtilsA1
  ;

type
    // TGLLODVectorFile
  TGLLODVectorFile = class(TVectorFile)
  public
    { Public Declarations }
    class function Capabilities : TDataFileCapabilities; override;

    procedure LoadFromStream(aStream : TStream); override;
  end;

var
    // Additional text information from LOD-file.
  LODFileTextInfo : string = '';

    // Physics
  EnablePhysicsLoading : Boolean = False;
  isPhysLoaded         : Boolean = False;
    // Main phisic arrays.
  LODBodys             : TLOD1Bodys;
  LODGeoms             : TLOD1Geoms;
  LODJoints            : TLOD1Joints;


implementation

  // Capabilities
  //
class function TGLLODVectorFile.Capabilities : TDataFileCapabilities;
begin
  Result := [dfcRead];
end;

  // LoadFromStream
  //
procedure TGLLODVectorFile.LoadFromStream(aStream : TStream);
var
  PHS1Chunk, 
  Inf1Chunk,
  MainChunk, Chunk2Level, MaterChunk,
  SkelChunk, AnimsChunk, Anim1Chunk  : TLODChunk;
  MeshChunk                          : array of TLODChunk;
  FirstMesh, mo1, i, c2, an1         : Integer;
  s1                                 : string;
  GropusEx1                          : TLODGroupsExs;
  isMeshLoaded, isSkeletonLoaded     : Boolean;
  LODAnimExDatas1                    : TLODAnimExDatas1;
begin
  if Owner = nil then exit;
  with aStream do begin
    try
      SetLength(s1, Length(LODkaSignature1));
      Read(s1[1],  Length(LODkaSignature1));
      if CompareText(s1, LODkaSignature1) <> 0 then begin
        Assert(False, 'Not valid *.LOD file.');
        exit;
      end;

      Read(MainChunk, SizeOf(TLODChunk));
      if MainChunk.ID <> 'LOD1' then begin
        Assert(False, 'Wrong version of *.LOD file loader. Try to download'
         +' a new version from LODka''s site');
        exit;
      end;

      isPhysLoaded     := False;
      LODFileTextInfo  := '';
      isSkeletonLoaded := False;
      isMeshLoaded     := False;
      FirstMesh        := Owner.MeshObjects.Count;
      for c2 := 0 to MainChunk.Count -1 do begin
        Read(Chunk2Level, SizeOf(TLODChunk));


          // Physics.
        if Chunk2Level.ID = 'PHS1' then begin
          PHS1Chunk := Chunk2Level;
          if     EnablePhysicsLoading
             and (PHS1Chunk.Count = 1)
          then begin
            LoadPhysicsFromStream1(aStream, LODBodys, LODGeoms, LODJoints);
            isPhysLoaded := True;
          end else begin
            Position := Position +PHS1Chunk.Size;
          end;
        end else

          // LODFileTextInfo.
        if Chunk2Level.ID = 'INF1' then begin
          Inf1Chunk := Chunk2Level;
          if Inf1Chunk.Count = 1 then begin
            SetLength(LODFileTextInfo, Inf1Chunk.Size);
            Read(LODFileTextInfo[1], Inf1Chunk.Size);
          end else begin
            Position := Position +Inf1Chunk.Size;
          end;
        end else

          // Skeleton.
        if Chunk2Level.ID = 'SKL1' then begin
          if not isMeshLoaded then begin
            SkelChunk := Chunk2Level;
            ReadSkeleton1(aStream, Owner, Chunk2Level.Count);
            isSkeletonLoaded := True;
          end else begin
            Position := Position +Chunk2Level.Size;
          end;
        end else

          // Animations.
        if Chunk2Level.ID = 'ANS1' then begin
          if     (Owner is TGLActor)
             and isSkeletonLoaded
          then begin
            AnimsChunk := Chunk2Level;
            SetLength(LODAnimExDatas1, AnimsChunk.Count);
            for an1 := 0 to AnimsChunk.Count -1 do begin
              Read(Anim1Chunk,  SizeOf(TLODChunk));
              ReadAnim1(aStream, Owner, an1, Anim1Chunk.Count,
                SkelChunk.Count
                , LODAnimExDatas1[an1]
              );
            end;
          end else begin
            Position := Position +Chunk2Level.Size;
          end;
        end else

        if Chunk2Level.ID = 'MAL1' then begin // Material Library
          Read(MaterChunk,     SizeOf(TLODChunk));
          ReadMat1(aStream, Owner.MaterialLibrary);
        end else

        if Chunk2Level.ID = 'MSL1' then begin // Mesh Library
          isMeshLoaded := True;
          SetLength(MeshChunk, Chunk2Level.Count);
          for mo1 := 0 to Chunk2Level.Count -1 do begin
            Read(MeshChunk[mo1], SizeOf(TLODChunk));
            ReadMesh1(aStream, Owner.MeshObjects, GropusEx1, isSkeletonLoaded,
              SkelChunk.Count);
          end;
        end else begin
          Position := Position +Chunk2Level.Size;
        end;
      end;

      if isSkeletonLoaded then begin
        Owner.Skeleton.RootBones.PrepareGlobalMatrices;
        for i := FirstMesh to Owner.MeshObjects.Count -1 do
          TSkeletonMeshObject(Owner.MeshObjects[i]).PrepareBoneMatrixInvertedMeshes;
      end;
    except
      Assert(False, 'Something wrong in GLFileLOD.pas .');
    end;
  end;
end;

initialization

  RegisterVectorFileFormat('lod', 'LODka 3d models editor files',
    TGLLODVectorFile);

end.