unit ULODkaUtilsA1;

{: ULODkaUtils<p>
   LODka 3d models editor files

     History:
   date  7 jun 2006 - Add physics support.
   date  1 may 2006 - Add skeleton support.
   date 23 feb 2006 - Creation.
}

interface

{$DEFINE UseSkeleton}

uses
  SysUtils, Classes, GLVectorFileObjects, GLTexture, VectorGeometry
  ;

const
  LODkaSignature1 = 'LODka3D1';

type
  TLODChunk = packed record
    ID    : array [0..3] of Char;
    Size  : Integer;
    Count : Integer;
  end;


    // Physics
  TBodyTypeA = ( bdtNone, bdtBox, bdtSphere, bdtCylinder, bdtCCylinder,
    bdtPlane, bdtTriMesh);

  TJointTypeA = (jtaNone, jtaBall, jtaSlider, jtaHinge, jtaHinge2,
    jtaUniversal);

  TLOD1Body = packed record
    Type1       : TBodyTypeA;
    Name        : string;
    Position    : TAffineVector;
    Scale       : TAffineVector;
    Rotation    : TQuaternion;
    Axis        : Integer;
    Mass        : Single;
    MassDensity : Single;
    TagFloat    : Single;
  end;
  TLOD1Bodys = array of TLOD1Body;

  TLOD1Geom = packed record
    Type1       : TBodyTypeA;
    Name        : string;
    BodyIndex   : Integer;
    Position    : TAffineVector;
    Scale       : TAffineVector;
    Rotation    : TQuaternion;
    Axis        : Integer;
    Mass        : Single;
    MassDensity : Single;
    TagFloat    : Single;
    MeshIndex   : array of Integer;
  end;
  TLOD1Geoms = array of TLOD1Geom;

  TLOD1Joint = packed record
    JointTypeA          : TJointTypeA;
    BodyIndex           : array [0..1] of Integer;
    Axis                : array [0..1] of TQuaternion;
    Position            : TAffineVector;
    dParamLoStop,       dParamHiStop, dParamVel,  dParamFMax,
    dParamFudgeFactor,  dParamBounce, dParamCFM,  dParamStopERP,
    dParamStopCFM,      dParamSuspensionERP,
    dParamSuspensionCFM, TagFloat : Single;
  end;
  TLOD1Joints = array of TLOD1Joint;






     // Animation
  TLODAnimationMode = (
    aamLODNone, aamLODPlayOnce, aamLODLoop, aamLODBounceForward,
    aamLODBounceBackward, aamLODLoopBackward, aamLODExternal);

  TLODAnimExData1 = packed record
    Interval : Integer;
    Mode     : TLODAnimationMode;
  end;

  TLODAnimExDatas1 = array of TLODAnimExData1;

  TLODSkelSave = (lodSSnone, lodSkSaveV1);

  TMaterialSaveType = (
    mstNotSave,
    mstOnlyNames,
    mstFourColors,
    mstTwoTextures
  );

    TMaterialSaveRecA = packed record
      MaterialName, FileName : string;
    end;

    TMaterialSaveRecAEx = packed record
      Ambient,  Diffuse,
      Emission, Specular, EnvColor   : TVector;
      Shininess                      : TShininess;
      ImageBrightness, ImageGamma,
      NormalMapScale                 : Single;
      BlendingMode                   : TBlendingMode;
      ImageAlpha                     : TGLTextureImageAlpha;
      TextureMode                    : TGLTextureMode;
      TextureWrap                    : TGLTextureWrap;
      TextureFormat                  : TGLTextureFormat;
      TextureEnabled                 : Boolean;
      Enabled                        : Boolean;
    end;

    TMaterialSaveRecAEx2 = packed record
      Ambient,  Diffuse,
      Emission, Specular, EnvColor   : TVector;
      Shininess                      : TShininess;
      ImageBrightness, ImageGamma,
      NormalMapScale                 : Single;
      BlendingMode                   : TBlendingMode;
      ImageAlpha                     : TGLTextureImageAlpha;
      TextureMode                    : TGLTextureMode;
      TextureWrap                    : TGLTextureWrap;
      TextureFormat                  : TGLTextureFormat;
      TextureEnabled                 : Boolean;
      Enabled                        : Boolean;
      MappingMode                    : TGLTextureMappingMode;
    end;

  TMaterialSaveRecM = packed record
    Names  : TMaterialSaveRecA;
    Params : TMaterialSaveRecAEx;
  end;

    TLODGroupsEx = packed record
      Name    : string;
      Enabled : Boolean;
    end;

  TLODGroupsExs  = array of TLODGroupsEx;
  TLODGroupsExss = array of TLODGroupsExs;

  function  ReadLODChunk(aStream : TStream; var aLODChunk : TLODChunk): Boolean;

  procedure ReadMat1(aStream : TStream; aMatLib : TGLMaterialLibrary);

  procedure ReadMesh1(aStream: TStream;
    aMeshObjects     : TMeshObjectList
    ; var aGropusEx  : TLODGroupsExs
    ; isLoadSkeleton : Boolean
    ; aBonesCount    : Integer = 0
  );

  procedure ReadSkeleton1(aStream : TStream; aBaseMesh : TGLBaseMesh
    ; BonesCount    : Integer
    );

  procedure ReadAnim1(aStream: TStream; aBaseMesh : TGLBaseMesh;
    anAnimNumber : Integer; aFramesCount : Integer
    ; aBonesCount : Integer
    ; aLODAnimExData : TLODAnimExData1
  );


    // Physics.
  function LoadPhysicsFromStream1(aStream : TStream;
    var aLODBodys  : TLOD1Bodys;
    var aLODGeoms  : TLOD1Geoms;
    var aLODJoints : TLOD1Joints
  ) : Boolean;

var
  isBuildNormalsIfNecessary : Boolean = True;

implementation

function InRange(const AValue, AMin, AMax: Integer): Boolean;
begin
  Result := (AValue >= AMin) and (AValue <= AMax);
end;

function ReadLODChunk(aStream: TStream; var aLODChunk : TLODChunk): Boolean;
begin
  Result := aStream.Position < aStream.Size;
  if Result then
    aStream.Read(aLODChunk, 12);
end;

procedure ReadMat1(aStream: TStream; aMatLib : TGLMaterialLibrary);
var
  MatSType   : TMaterialSaveType;
  MatSave    : array of TMaterialSaveRecA;
  MatSaveEx1 : array of TMaterialSaveRecAEx;
  MatSaveEx2 : array of TMaterialSaveRecAEx2;

  procedure PutArraysToMatLib;
  var
    ps1,
    i      : Integer;
    MatLib : TGLMaterialLibrary;
    LbMat  : TGLLibMaterial;
    sFileName, sTextrue2Name : string;
  begin
    MatLib := aMatLib;
    if MatLib <> nil then
      for i := 0 to Length(MatSave) -1 do begin
        ps1 := Pos(':', MatSave[i].FileName);
        if ps1 = 0 then begin
          sFileName     := MatSave[i].FileName;
          sTextrue2Name := '';
        end else begin
          sFileName     := Copy(MatSave[i].FileName, 1, ps1 -1);
          sTextrue2Name := Copy(MatSave[i].FileName, ps1 +1, MaxInt);
        end;
        LbMat := MatLib.Materials.GetLibMaterialByName(
          MatSave[i].MaterialName);
        if LbMat = nil then begin
          if FileExists(sFileName) then begin
            LbMat := MatLib.AddTextureMaterial(MatSave[i].MaterialName,
              sFileName);
            LbMat.Texture2Name := sTextrue2Name;
          end else begin
            LbMat      := MatLib.Materials.Add;
            LbMat.Name := MatSave[i].MaterialName;
            LbMat.Texture2Name := sTextrue2Name;
          end;
        end else begin
          if FileExists(sFileName) then begin
            LbMat.Material.Texture.Image.LoadFromFile(
              sFileName);
            LbMat.Material.Texture.Disabled := False;
          end;
          LbMat.Texture2Name := sTextrue2Name;
        end;
      end;
  end;

  procedure PutArraysToMatLibEx;
  var
    k2, i : Integer;
    LbMat : TGLLibMaterial;
  begin
    if aMatLib = nil then exit;
    k2 := Length(MatSaveEx1);
    if Length(MatSave) < k2 then begin
      beep;
      k2 := Length(MatSave);
    end;
    for i := 0 to k2 -1 do begin
      if i > Length(MatSaveEx2) -1 then begin
        // beep;
      end;
      if not MatSaveEx1[i].Enabled then continue;
      LbMat := aMatLib.Materials.GetLibMaterialByName(
        MatSave[i].MaterialName);
      if LbMat <> nil then begin
        with LbMat.Material do begin
          BlendingMode                    := MatSaveEx1[i].BlendingMode;
          FrontProperties.Ambient.Color   := MatSaveEx1[i].Ambient;
          FrontProperties.Diffuse.Color   := MatSaveEx1[i].Diffuse;
          FrontProperties.Emission.Color  := MatSaveEx1[i].Emission;
          FrontProperties.Specular.Color  := MatSaveEx1[i].Specular;
          FrontProperties.Shininess       := MatSaveEx1[i].Shininess;
          Texture.Disabled                := not MatSaveEx1[i].TextureEnabled;
          Texture.EnvColor.Color          := MatSaveEx1[i].EnvColor;
          Texture.ImageBrightness         := MatSaveEx1[i].ImageBrightness;
          Texture.ImageGamma              := MatSaveEx1[i].ImageGamma;
          Texture.NormalMapScale          := MatSaveEx1[i].NormalMapScale;
          Texture.ImageAlpha              := MatSaveEx1[i].ImageAlpha;
          Texture.TextureMode             := MatSaveEx1[i].TextureMode;
          Texture.TextureWrap             := MatSaveEx1[i].TextureWrap;
          Texture.TextureFormat           := MatSaveEx1[i].TextureFormat;
        end;
      end;
    end;
  end;

  procedure PutArraysToMatLibEx2;
  var
    k2, i : Integer;
    LbMat : TGLLibMaterial;
  begin
    if aMatLib = nil then exit;
    k2 := Length(MatSaveEx2);
    if Length(MatSave) < k2 then begin
      beep;
      k2 := Length(MatSave);
    end;
    for i := 0 to k2 -1 do begin
      if i > Length(MatSaveEx2) -1 then begin
        beep;
      end;
      if not MatSaveEx2[i].Enabled then continue;
      LbMat := aMatLib.Materials.GetLibMaterialByName(
        MatSave[i].MaterialName);
      if LbMat <> nil then begin
        with LbMat.Material do begin
          BlendingMode                    := MatSaveEx2[i].BlendingMode;
          FrontProperties.Ambient.Color   := MatSaveEx2[i].Ambient;
          FrontProperties.Diffuse.Color   := MatSaveEx2[i].Diffuse;
          FrontProperties.Emission.Color  := MatSaveEx2[i].Emission;
          FrontProperties.Specular.Color  := MatSaveEx2[i].Specular;
          FrontProperties.Shininess       := MatSaveEx2[i].Shininess;
          Texture.Disabled                := not MatSaveEx2[i].TextureEnabled;
          Texture.EnvColor.Color          := MatSaveEx2[i].EnvColor;
          Texture.ImageBrightness         := MatSaveEx2[i].ImageBrightness;
          Texture.ImageGamma              := MatSaveEx2[i].ImageGamma;
          Texture.NormalMapScale          := MatSaveEx2[i].NormalMapScale;
          Texture.ImageAlpha              := MatSaveEx2[i].ImageAlpha;
          Texture.TextureMode             := MatSaveEx2[i].TextureMode;
          Texture.TextureWrap             := MatSaveEx2[i].TextureWrap;
          Texture.TextureFormat           := MatSaveEx2[i].TextureFormat;
          Texture.MappingMode             := MatSaveEx2[i].MappingMode;
        end;
      end;
    end;
  end;

var
  i, MatCnt, kLen2 : Integer;
begin
  with aStream do begin
    Read(MatSType, SizeOf(TMaterialSaveType));
    if Integer(MatSType) > Integer(mstTwoTextures) then begin
      Assert(False, 'Wrong material type. Download a new version of ULODkaUtils file.');
    end;
    case MatSType of
      mstNotSave : begin
        end;
      mstOnlyNames : begin
          Read(MatCnt, SizeOf(Integer));
          SetLength(MatSave, MatCnt);
          for i := 0 to MatCnt -1 do begin
            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].MaterialName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].MaterialName[1], kLen2);

            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].FileName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].FileName[1], kLen2);
          end;
          PutArraysToMatLib;
        end;
      mstFourColors : begin
          Read(MatCnt, SizeOf(Integer));
          SetLength(MatSave,    MatCnt);
          SetLength(MatSaveEx1, MatCnt);
          for i := 0 to MatCnt -1 do begin
            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].MaterialName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].MaterialName[1], kLen2);
            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].FileName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].FileName[1], kLen2);

            kLen2 := SizeOf(TMaterialSaveRecAEx);
            if kLen2 > 0 then
              Read(MatSaveEx1[i], kLen2);
          end;
          PutArraysToMatLib;
          PutArraysToMatLibEx;
        end;

      mstTwoTextures : begin
          Read(MatCnt, SizeOf(Integer));
          SetLength(MatSave,    MatCnt);
          SetLength(MatSaveEx2, MatCnt);
          for i := 0 to MatCnt -1 do begin
            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].MaterialName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].MaterialName[1], kLen2);

            Read(kLen2, SizeOf(Integer));
            SetLength(MatSave[i].FileName, kLen2);
            if kLen2 > 0 then
              Read(MatSave[i].FileName[1], kLen2);

            kLen2 := SizeOf(TMaterialSaveRecAEx2);
            if kLen2 > 0 then
              Read(MatSaveEx2[i], kLen2);
          end;
          PutArraysToMatLib;
          PutArraysToMatLibEx2;
        end;
    end;
  end;
end;


procedure ReadMesh1(aStream: TStream;
  aMeshObjects     : TMeshObjectList
  ; var aGropusEx  : TLODGroupsExs
  ; isLoadSkeleton : Boolean
  ; aBonesCount    : Integer = 0
  );
type
  TTwoSingle = array [0..1] of Single;
var
  mMode1          : TMeshObjectMode;
  fMode1          : TFaceGroupMeshMode;
  vertCnt, normCnt,
  BoneCnt, BoneWidthCnt,
  txcrCnt, n1,
  i, kCnt, fg1    : Integer;
  isSkeletonExists,
  isVisible       : Boolean;
  af2             : array of TAffineVector;
  sMName, sFName  : string;
  Mesh            : TMeshObject;
  SMeshO          : TSkeletonMeshObject;
  av1             : TAffineVector;

  aIndI       : array of Integer;
  aIndW       : array of Word;
  aIndB       : array of Byte;
  aInd2S      : array of TTwoSingle;

  {$IFDEF UseSkeleton}
  BnWidB      : array of Byte;
  BnWidW      : array of Word;
  {$ENDIF}
begin
  with aStream do begin

    aStream.Read(kCnt, SizeOf(Integer));
    if kCnt > 0 then begin
      SetLength(sMName, kCnt);
      if kCnt > 0 then
        aStream.Read(sMName[1], kCnt);
    end;

    aStream.Read(isSkeletonExists, SizeOf(Boolean));

    if isLoadSkeleton then begin
      SMeshO := TSkeletonMeshObject.CreateOwned(aMeshObjects);
      Mesh   := SMeshO;
    end else begin
      Mesh   := TMeshObject.CreateOwned(aMeshObjects)
    end;

    with Mesh do begin
      Name := sMName;

      aStream.Read(mMode1, SizeOf(TMeshObjectMode));
      Mode := mMode1;

      aStream.Read(isVisible, SizeOf(Boolean));
      Visible := isVisible;

      aStream.Read(vertCnt, SizeOf(Integer));
      if vertCnt > 0 then begin
        SetLength(af2, vertCnt);
        aStream.Read(af2[0], vertCnt *SizeOf(TAffineVector));
        for i := 0 to vertCnt -1 do
          Vertices.Add(af2[i]);
      end;

      aStream.Read(normCnt, SizeOf(Integer));
      if isLoadSkeleton then begin
        aStream.Position := aStream.Position +(normCnt *SizeOf(TAffineVector));
        for i := 0 to vertCnt -1 do
          Normals.Add(NullVector);
      end else

      begin
        if normCnt > 0 then begin
          SetLength(af2, normCnt);
          aStream.Read(af2[0], normCnt *SizeOf(TAffineVector));
          for i := 0 to normCnt -1 do
            Normals.Add(af2[i]);
        end;
      end;

      aStream.Read(txcrCnt, SizeOf(Integer));
      if txcrCnt > 0 then begin
        SetLength(aInd2S, txcrCnt);
        aStream.Read(aInd2S[0], txcrCnt *SizeOf(TTwoSingle));
        for i := 0 to txcrCnt -1 do
          TexCoords.Add(aInd2S[i][0], aInd2S[i][1], 0);
      end;

      {$IFDEF UseSkeleton}
      aStream.Read(BoneCnt, SizeOf(Integer));
      if BoneCnt > 0 then begin
        if isLoadSkeleton then begin
          if aBonesCount < 255 then begin
            SetLength(BnWidB, BoneCnt);
            aStream.Read(BnWidB[0], BoneCnt *SizeOf(Byte));
            for i := 0 to BoneCnt -1 do
              if InRange(BnWidB[i], 0, aBonesCount -1)
                then SMeshO.AddWeightedBone(BnWidB[i], 1)
                else SMeshO.AddWeightedBone(0, 1);
          end else begin
            SetLength(BnWidW, BoneCnt);
            aStream.Read(BnWidW[0], BoneCnt *SizeOf(Word));
            for i := 0 to BoneCnt -1 do
              if InRange(BnWidW[i], 0, aBonesCount -1)
                then SMeshO.AddWeightedBone(BnWidW[i], 1)
                else SMeshO.AddWeightedBone(0, 1);
          end;
        end else begin
          if aBonesCount < 255 then begin
            Position := Position +BoneCnt *SizeOf(Byte);
          end else begin
            Position := Position +BoneCnt *SizeOf(Word);
          end;
        end;
      end;
      {$ELSE}
      aStream.Read(BoneCnt, SizeOf(Integer));
      {$ENDIF}

      aStream.Read(BoneWidthCnt, SizeOf(Integer));

      aStream.Read(kCnt, SizeOf(Integer));
      SetLength(aGropusEx, kCnt);
      for fg1 := 0 to kCnt -1 do begin
        aStream.Read(kCnt, SizeOf(Integer));
        if kCnt <> 1 then begin
          beep;
          Assert(True, 'Unknown FaceGroups format');
        end;

        with TFGVertexNormalTexIndexList.CreateOwned(FaceGroups) do begin
          aStream.Read(fMode1, SizeOf(TFaceGroupMeshMode));
          Mode := fMode1;

          aStream.Read(kCnt, SizeOf(Integer));
          if kCnt > 0 then begin
            SetLength(sFName, kCnt);
            aStream.Read(sFName[1], kCnt);
            aGropusEx[fg1].Name := sFName;
          end;
          aStream.Read(isVisible, SizeOf(Boolean));
          aGropusEx[fg1].Enabled := isVisible;

          aStream.Read(kCnt, SizeOf(Integer));
          SetLength(sFName, kCnt);
          if kCnt > 0 then
            aStream.Read(sFName[1], kCnt);
          MaterialName := sFName;

          aStream.Read(kCnt, SizeOf(Integer));

          if kCnt > 0 then
            if vertCnt <= High(Byte) then begin
              SetLength(aIndB, kCnt);
              aStream.Read(aIndB[0], SizeOf(Byte) *kCnt);
              for i := 0 to kCnt -1 do
                VertexIndices.Add(aIndB[i]);
            end else if vertCnt <= High(Word) then begin
              SetLength(aIndW, kCnt);
              aStream.Read(aIndW[0], SizeOf(Word) *kCnt);
              for i := 0 to kCnt -1 do
                VertexIndices.Add(aIndW[i]);
            end else begin
              SetLength(aIndI, kCnt);
              aStream.Read(aIndI[0], SizeOf(Integer) *kCnt);
              for i := 0 to kCnt -1 do
                VertexIndices.Add(aIndI[i]);
            end;

          aStream.Read(kCnt, SizeOf(Integer));
          if kCnt > 0 then
            if normCnt <= High(Byte) then begin
              SetLength(aIndB, kCnt);
              aStream.Read(aIndB[0], SizeOf(Byte) *kCnt);
              for i := 0 to kCnt -1 do
                NormalIndices.Add(aIndB[i]);
            end else if normCnt <= High(Word) then begin
              SetLength(aIndW, kCnt);
              aStream.Read(aIndW[0], SizeOf(Word) *kCnt);
              for i := 0 to kCnt -1 do
                NormalIndices.Add(aIndW[i]);
            end else begin
              SetLength(aIndI, kCnt);
              aStream.Read(aIndI[0], SizeOf(Integer) *kCnt);
              for i := 0 to kCnt -1 do
                NormalIndices.Add(aIndI[i]);
            end;

          if isLoadSkeleton then begin
            NormalIndices.Clear;
            for i := 0 to (VertexIndices.Count div 3) -1 do begin
              av1 := CalcPlaneNormal(
                Vertices[ VertexIndices[i *3 +0] ],
                Vertices[ VertexIndices[i *3 +1] ],
                Vertices[ VertexIndices[i *3 +2] ]  );

              Normals.TranslateItem(VertexIndices[i *3 +0], av1);
              Normals.TranslateItem(VertexIndices[i *3 +1], av1);
              Normals.TranslateItem(VertexIndices[i *3 +2], av1);

              NormalIndices.Add(VertexIndices[i *3 +0]);
              NormalIndices.Add(VertexIndices[i *3 +1]);
              NormalIndices.Add(VertexIndices[i *3 +2]);
            end;
            if VertexIndices.Count <> NormalIndices.Count then begin
              beep;
            end;
          end else
          begin

            if     (kCnt    = 0)
               and (normCnt = 0)
               and isBuildNormalsIfNecessary
            then begin
              for i := 0 to (VertexIndices.Count div 3) -1 do begin

                av1 := CalcPlaneNormal( Vertices[ VertexIndices[i *3 +0] ],
                                        Vertices[ VertexIndices[i *3 +1] ],
                                        Vertices[ VertexIndices[i *3 +2] ]  );
                n1 := Normals.FindOrAdd(av1);

                NormalIndices.Add(n1);
                NormalIndices.Add(n1);
                NormalIndices.Add(n1);
              end;
              if VertexIndices.Count <> NormalIndices.Count then begin
                beep;
              end;
            end;
          end;

          aStream.Read(kCnt, SizeOf(Integer));
          if kCnt > 0 then
            if txcrCnt <= High(Byte) then begin
              SetLength(aIndB, kCnt);
              aStream.Read(aIndB[0], SizeOf(Byte) *kCnt);
              for i := 0 to kCnt -1 do
                TexCoordIndices.Add(aIndB[i]);
            end else if txcrCnt <= High(Word) then begin
              SetLength(aIndW, kCnt);
              aStream.Read(aIndW[0], SizeOf(Word) *kCnt);
              for i := 0 to kCnt -1 do
                TexCoordIndices.Add(aIndW[i]);
            end else begin
              SetLength(aIndI, kCnt);
              aStream.Read(aIndI[0], SizeOf(Integer) *kCnt);
              for i := 0 to kCnt -1 do
                TexCoordIndices.Add(aIndI[i]);
            end;
        end;
      end;
    end;

  end;
end;



procedure ReadSkeleton1(aStream : TStream; aBaseMesh : TGLBaseMesh
  ; BonesCount    : Integer
  );
type
  TLODSkelSave = (lodSSnone, lodSkSaveV1);

    TSkelBone = record
      Name         : string;
      ID, ParentID : Smallint;
    end;
  TSkelBones = array of TSkelBone;

var
  Bones : TSkelBones;

  procedure WriteBonesToActor(aSkeletone : TSkeleton);
  var
    Bone1, BoneParent : TSkeletonBone;
    i                 : integer;
  begin
    if aSkeletone.RootBones.Count = 0 then begin
      for i := 0 to Length(Bones) -1 do begin
        if Bones[i].ParentID = -1 then begin
          Bone1 := TSkeletonBone.CreateOwned( aSkeletone.RootBones);
        end else begin
          BoneParent := aSkeletone.RootBones.BoneByID(Bones[i].ParentID);
          if Assigned(BoneParent) then begin
            Bone1 := TSkeletonBone.CreateOwned( BoneParent );
          end else begin
            Assert(False, 'Error ! Parent bone number ( '
             +IntToStr(Bones[i].ParentID) +' ) is not valid.');
             continue;
          end;
        end;
        if Assigned(Bone1) then begin
          Bone1.Name   := Bones[i].Name;
          Bone1.BoneID := Bones[i].ID;
        end;
      end;
    end else begin
      Assert(False, 'Error ! Skeletone is already exisxts');
    end;
  end;

var
  i, k1     : integer;
  SkelType2 : TLODSkelSave;
begin
  with aStream do begin
    Read(SkelType2, SizeOf(TLODSkelSave));
    if Integer(SkelType2) > Integer(lodSkSaveV1) then begin
      Assert(False, 'Wrong skeletone type. Download a new version of the'
        +' loader from LODka3D site.');
      exit;
    end;
    SetLength(Bones, BonesCount);
    for i := 0 to BonesCount -1 do begin
      Read(Bones[i].ID, SizeOf(SmallInt));
      Read(Bones[i].ParentID, SizeOf(SmallInt));
      Read(k1, SizeOf(Integer));
      SetLength(Bones[i].Name, k1);
      Read(Bones[i].Name[1], k1);
    end;
  end;
  WriteBonesToActor(aBaseMesh.Skeleton);
end;


procedure ReadAnim1(aStream: TStream; aBaseMesh : TGLBaseMesh;
  anAnimNumber : Integer; aFramesCount : Integer
  ; aBonesCount : Integer
  ; aLODAnimExData : TLODAnimExData1
  );
type
  TSkelPos2 = record
    Position, Rotation : TAffineVector;
  end;
  TSkelAnim2 = record
    Name  : string;
    Items : array of TSkelPos2;
  end;
  TLODAnimType1 = (lodATnone, lodATSaveV1);
var
  Actor     : TGLActor;
  Anim1     : TSkelAnim2;
  AnimType1 : TLODAnimType1;
  ActAnim   : TActorAnimation;

  k1, firstFrame, BonesCnt,
  fr1, BoneNumb             : Integer;
  Frame1                    : TSkeletonFrame;
begin
  if not (aBaseMesh is TGLActor) then begin
    exit;
  end;
  Actor := aBaseMesh as TGLActor;
  BonesCnt := aBonesCount;
  with aStream do begin
    AnimType1 := lodATSaveV1;
    Read(AnimType1, SizeOf(TLODAnimType1));
    if AnimType1 <> lodATSaveV1 then begin
      Assert(False,
        'Wrong skeleton type. Download a new version of ULODkaUtils file.');
      exit;
    end;

    Read(aLODAnimExData, SizeOf(TLODAnimExData1));

    Read(k1, SizeOf(Integer));
    SetLength(Anim1.Name, k1);
    Read(Anim1.Name[1], k1);

    SetLength(Anim1.Items, aFramesCount *BonesCnt);
    if aFramesCount *BonesCnt > 0 then
      Read(Anim1.Items[0], SizeOf(TAffineVector) *2 *aFramesCount *BonesCnt)
    else begin
      beep;
      exit;
    end;
  end;

  FirstFrame   := Actor.Skeleton.Frames.Count;
  k1 := 0;
  for fr1 := 0 to aFramesCount -1 do begin
    Frame1      := TSkeletonFrame.CreateOwned(Actor.Skeleton.Frames);
    Frame1.Name := 'a' +IntToStr(fr1 +FirstFrame);
    for BoneNumb := 0 to BonesCnt -1 do begin
      Frame1.Position.Add(Anim1.Items[k1].Position);
      Frame1.Rotation.Add(Anim1.Items[k1].Rotation);
      k1 := k1 +1;
    end;
  end;

  ActAnim            := Actor.Animations.Add;
  ActAnim.Name       := Anim1.Name;
  ActAnim.Reference  := aarSkeleton;
  ActAnim.StartFrame := FirstFrame;
  ActAnim.EndFrame   := Actor.Skeleton.Frames.Count -1;
end;











  // Physics.
function LoadPhysicsFromStream1(aStream : TStream;
  var aLODBodys  : TLOD1Bodys;
  var aLODGeoms  : TLOD1Geoms;
  var aLODJoints : TLOD1Joints
) : Boolean;
type
    // Temp records.
  TSaveLens = packed record
    BnLen, GnLen, MsLen, BdLen, GmLen, JnLen : Integer;
  end;

  TLOD01Body = packed record
    Type1       : TBodyTypeA;
    Position    : TAffineVector;
    Scale       : TAffineVector;
    Rotation    : TQuaternion;
    Axis        : Integer;
    Mass        : Single;
    MassDensity : Single;
    TagFloat    : Single;
  end;
  TLOD01Bodys = array of TLOD01Body;

  TLOD01Geom = packed record
    Type1       : TBodyTypeA;
    BodyIndex   : Integer;
    Position    : TAffineVector;
    Scale       : TAffineVector;
    Rotation    : TQuaternion;
    Axis        : Integer;
    Mass        : Single;
    MassDensity : Single;
    MeshLink,
    MeshCount   : Integer;
    TagFloat    : Single;
  end;
  TLOD01Geoms = array of TLOD01Geom;

  TLOD01BodyGeomsName = string;

  TLOD1MeshLinksList = array of Integer;

    procedure MakePhysRecs(
      Bodys01           : TLOD01Bodys;
      Geoms01           : TLOD01Geoms;
      Joints01          : TLOD1Joints;

      BodysNames01      : TLOD01BodyGeomsName;
      GeomsNames01      : TLOD01BodyGeomsName;
      MeshLinks01       : TLOD1MeshLinksList;

      var aLOD1Bodys02  : TLOD1Bodys;
      var aLOD1Geoms02  : TLOD1Geoms;
      var aLOD1Joints02 : TLOD1Joints
      );
    var
      i, b1, g1, j1, k1, j2  : Integer;
      BNamesList, GNamesList : TStringList;
    begin
      b1 := Length(Bodys01);
      SetLength(aLOD1Bodys02,  b1);
      g1 := Length(Geoms01);
      SetLength(aLOD1Geoms02,  g1);
      j1 := Length(Joints01);
      SetLength(aLOD1Joints02, j1);

      BNamesList := TStringList.Create;
      GNamesList := TStringList.Create;
      try
        BNamesList.Text := BodysNames01;
        GNamesList.Text := GeomsNames01;

        for i := 0 to b1 -1 do begin
          aLOD1Bodys02[i].Type1 := Bodys01[i].Type1;
          if i <= BNamesList.Count -1
            then aLOD1Bodys02[i].Name := BNamesList[i]
            else aLOD1Bodys02[i].Name := '';
          aLOD1Bodys02[i].Position    := Bodys01[i].Position;
          aLOD1Bodys02[i].Scale       := Bodys01[i].Scale;
          aLOD1Bodys02[i].Rotation    := Bodys01[i].Rotation;
          aLOD1Bodys02[i].Axis        := Bodys01[i].Axis;
          aLOD1Bodys02[i].Mass        := Bodys01[i].Mass;
          aLOD1Bodys02[i].MassDensity := Bodys01[i].MassDensity;
          aLOD1Bodys02[i].TagFloat    := Bodys01[i].TagFloat;
        end;

        for i := 0 to g1 -1 do begin
          aLOD1Geoms02[i].Type1 := Geoms01[i].Type1;
          if i <= GNamesList.Count -1
            then aLOD1Geoms02[i].Name := GNamesList[i]
            else aLOD1Geoms02[i].Name := '';
          aLOD1Geoms02[i].BodyIndex   := Geoms01[i].BodyIndex;
          aLOD1Geoms02[i].Position    := Geoms01[i].Position;
          aLOD1Geoms02[i].Scale       := Geoms01[i].Scale;
          aLOD1Geoms02[i].Rotation    := Geoms01[i].Rotation;
          aLOD1Geoms02[i].Axis        := Geoms01[i].Axis;
          aLOD1Geoms02[i].Mass        := Geoms01[i].Mass;
          aLOD1Geoms02[i].MassDensity := Geoms01[i].MassDensity;
          aLOD1Geoms02[i].TagFloat    := Geoms01[i].TagFloat;
          for j2 := 0 to Geoms01[i].MeshCount -1 do
            if InRange(Geoms01[i].MeshLink +j2, 0, Length(MeshLinks01) -1) then begin
              k1 := Length(aLOD1Geoms02[i].MeshIndex);
              SetLength(aLOD1Geoms02[i].MeshIndex, k1 +1);
              aLOD1Geoms02[i].MeshIndex[k1] := MeshLinks01[ Geoms01[i].MeshLink + j2];
            end;
        end;

        for i := 0 to j1 -1 do begin
          aLOD1Joints02[i].JointTypeA := Joints01[i].JointTypeA;
          for k1 := 0 to 1 do begin
            aLOD1Joints02[i].BodyIndex[k1] := Joints01[i].BodyIndex[k1];
            aLOD1Joints02[i].Axis[k1]      := Joints01[i].Axis[k1];
          end;
          aLOD1Joints02[i].Position            := Joints01[i].Position;
          aLOD1Joints02[i].dParamLoStop        := Joints01[i].dParamLoStop;
          aLOD1Joints02[i].dParamHiStop        := Joints01[i].dParamHiStop;
          aLOD1Joints02[i].dParamVel           := Joints01[i].dParamVel;
          aLOD1Joints02[i].dParamFMax          := Joints01[i].dParamFMax;
          aLOD1Joints02[i].dParamFudgeFactor   := Joints01[i].dParamFudgeFactor;
          aLOD1Joints02[i].dParamBounce        := Joints01[i].dParamBounce;
          aLOD1Joints02[i].dParamCFM           := Joints01[i].dParamCFM;
          aLOD1Joints02[i].dParamStopERP       := Joints01[i].dParamStopERP;
          aLOD1Joints02[i].dParamStopCFM       := Joints01[i].dParamStopCFM;
          aLOD1Joints02[i].dParamSuspensionERP := Joints01[i].dParamSuspensionERP;
          aLOD1Joints02[i].dParamSuspensionCFM := Joints01[i].dParamSuspensionCFM;
          aLOD1Joints02[i].TagFloat            := Joints01[i].TagFloat;
        end;

      finally
        BNamesList.Free;
        GNamesList.Free;
      end;
    end;

var
  SaveType1   : Byte;
  Bodys1      : TLOD01Bodys;
  Geoms1      : TLOD01Geoms;
  Joints1     : TLOD1Joints;
  BodysNames1 : TLOD01BodyGeomsName;
  GeomsNames1 : TLOD01BodyGeomsName;
  MeshLinks1  : TLOD1MeshLinksList;
  Lens1       : TSaveLens;
  b1          : Boolean;
begin
  Result := False;
  with aStream do begin
    try
        // Type.
      Read(SaveType1, SizeOf(SaveType1));
      if SaveType1 <> 1 then begin
        Result := False;
        exit;
      end;

      Read(Lens1, SizeOf(TSaveLens));
      SetLength(BodysNames1, Lens1.BnLen);
      SetLength(GeomsNames1, Lens1.GnLen);
      SetLength(MeshLinks1,  Lens1.MsLen);
      SetLength(Bodys1,      Lens1.BdLen);
      SetLength(Geoms1,      Lens1.GmLen);
      SetLength(Joints1,     Lens1.JnLen);
      if Lens1.BnLen > 0 then
        Read(BodysNames1[1], Lens1.BnLen);
      if Lens1.GnLen > 0 then
        Read(GeomsNames1[1], Lens1.GnLen);
      if Lens1.MsLen > 0 then
        Read(MeshLinks1[0],  Lens1.MsLen *SizeOf(Integer));
      if Lens1.BdLen > 0 then
        Read(Bodys1[0],  Lens1.BdLen *SizeOf(TLOD01Body));
      if Lens1.GmLen > 0 then
        Read(Geoms1[0],  Lens1.GmLen *SizeOf(TLOD01Geom));
      if Lens1.JnLen > 0 then
        Read(Joints1[0], Lens1.JnLen *SizeOf(TLOD1Joint));
      MakePhysRecs(
        Bodys1, Geoms1, Joints1,
        BodysNames1, GeomsNames1, MeshLinks1,
        aLODBodys, aLODGeoms, aLODJoints );
      Result := True;
    except
      Result := False;
    end;
  end;
end;

end.