unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  GLScene, GLObjects, GLCadencer, GLMisc, GLVectorFileObjects, ExtCtrls,
  StdCtrls, GLWin32Viewer, GLTexture, VectorGeometry

  , dynode, dynodegl // Use ODE.
  ;

type
  TForm1 = class(TForm)
    Viewer: TGLSceneViewer;
    GLScene: TGLScene;
    GLCadencer: TGLCadencer;
    GLCamera1: TGLCamera;
    GLLightSource1: TGLLightSource;
    DCCameraTarget: TGLDummyCube;
    Timer1: TTimer;
    GLMaterialLibrary: TGLMaterialLibrary;
    GLLightSource2: TGLLightSource;
    GLFreeForm1: TGLFreeForm;
    Panel1: TPanel;
    DCODE1: TGLDummyCube;
    Button1: TButton;
    Label1: TLabel;
    procedure GLCadencerProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure Timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ViewerMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure ViewerMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);

    procedure CreateODE1;
    procedure DestroyODE1;
    procedure UpdateODE1;
    procedure Button1Click(Sender: TObject);
  private
    mdx, mdy : Integer;
  public
  end;

  procedure nearCallback(Data : Pointer; O1, O2 : PdxGeom); cdecl;

var
  Form1: TForm1;

    // Physics.
  World1        : PdxWorld;        // ode world
  Space1        : PdxSpace;        // ode space
  ContactGroup1 : TdJointGroupID;  // ode contactgroup
  ContactNum1   : Integer;         // ode contact numbers
  PhysTime      : Double;          // ode time update

  GeomsList     : array of PdxGeom; // Geoms list.


implementation

{$R *.DFM}

uses GLFileLOD, ULODkaUtilsA1;        // 1) Use LOD-file.


function MatrixToODERotation(Mat : TMatrix): TdMatrix3;
begin
  Result[0]  := Mat[0][0];
  Result[1]  := Mat[1][0];
  Result[2]  := Mat[2][0];
  Result[3]  := 0;
  Result[4]  := Mat[0][1];
  Result[5]  := Mat[1][1];
  Result[6]  := Mat[2][1];
  Result[7]  := 0;
  Result[8]  := Mat[0][2];
  Result[9]  := Mat[1][2];
  Result[10] := Mat[2][2];
  Result[11] := 0;
end;


procedure TForm1.FormCreate(Sender: TObject);
var
  i, bd1, gm1, gm2 : Integer;
  Body01           : PdxBody;
  Cb1              : TGLCube;
  ff1              : TGLFreeForm;
  Mesh             : TMeshObject;
begin
    // 2) Enable physic loading.
  EnablePhysicsLoading := True;
    // 3) Load LOD-file.
  GLFreeForm1.LoadFromFile('ode_world_01.lod');

    // 4) Read 3 physics array ( LODBodys, LODGeoms, LODJoints ) and build
    //    an ODE objects.

  for i := 0 to GLFreeForm1.MeshObjects.Count -1 do
    GLFreeForm1.MeshObjects[i].Visible := True;
  GLFreeForm1.Visible := False;

  Label1.Caption := Format('Bodys:%d Geoms:%d Joints:%d', [
    Length(LODBodys),
    Length(LODGeoms),
    Length(LODJoints)]);

  CreateODE1;

  for bd1 := 0 to Length(LODBodys) -1 do begin
      // Body.
    if CompareText(LODBodys[bd1].Name, 'World') = 0 then begin
      Body01 := nil;
    end else begin
      Body01 := dBodyCreate(World1);
      dBodySetPosition(Body01, LODBodys[bd1].Position[0],
                               LODBodys[bd1].Position[1],
                               LODBodys[bd1].Position[2]);
      dBodySetRotation(Body01,
        MatrixToODERotation( QuaternionToMatrix(LODBodys[bd1].Rotation) ) );

        // Mass.
      dMassSetZero(Body01.Mass);
      dMassSetBox(Body01.Mass, LODBodys[bd1].MassDensity,
        LODBodys[bd1].Scale[0],
        LODBodys[bd1].Scale[0],
        LODBodys[bd1].Scale[0] );
      dBodySetMass(Body01, @Body01.Mass);
    end;
      // Create geoms.
    for gm1 := 0 to Length(LODGeoms) -1 do begin
      if LODGeoms[gm1].BodyIndex <> bd1 then continue;
      if LODGeoms[gm1].Type1 <> bdtBox then continue;

        // Graphics.
      FF1 := nil;
      if     (Length(LODGeoms[gm1].MeshIndex) >  0)
         and (LODGeoms[gm1].MeshIndex[0] >= 0)
         and (LODGeoms[gm1].MeshIndex[0] <= GLFreeForm1.MeshObjects.Count -1)
      then begin
        FF1 := TGLFreeForm(DCODE1.AddNewChild(TGLFreeForm));
        with FF1 do begin
          FF1.MaterialLibrary := GLFreeForm1.MaterialLibrary;
          Mesh := TMeshObject.CreateOwned(FF1.MeshObjects);
          Mesh.Assign(GLFreeForm1.MeshObjects[ LODGeoms[gm1].MeshIndex[0] ]);
        end;
      end;
      gm2 := Length(GeomsList);
      SetLength(GeomsList, gm2 +1);
      GeomsList[gm2] := dCreateBox(Space1,
        LODGeoms[gm1].Scale[0],
        LODGeoms[gm1].Scale[1],
        LODGeoms[gm1].Scale[2]);
        // Set to body.
      dGeomSetBody(GeomsList[gm2], Body01);
        // Set to graphic.
      GeomsList[gm2].Data := FF1;
      break;
    end;
  end;
  UpdateODE1;
end;


procedure TForm1.FormDestroy(Sender: TObject);
begin
  DestroyODE1;
end;


procedure TForm1.GLCadencerProgress(Sender: TObject; const deltaTime,
  newTime: Double);
begin
  while physTime < newtime do
    UpdateODE1;

  Viewer.Invalidate;
end;


  // Start button.
procedure TForm1.Button1Click(Sender: TObject);
begin
  Button1.Enabled    := False;
  physTime           := GLCadencer.CurrentTime;
  GLCadencer.Enabled := True;
end;




  // --------------------------- ODE service ----------------------------------

procedure TForm1.UpdateODE1;
const
  PhysStep = 1 / 50;
var
  g1 : Integer;
begin
  physTime := physTime +PhysStep/8;

  dSpaceCollide(Space1, nil, nearCallback);
  dWorldQuickStep(World1, PhysStep);

  for g1 := 0 to Length(GeomsList) -1 do
    if     (GeomsList[g1]      <> nil)
       and (GeomsList[g1].Data <> nil)
    then
      PositionSceneObject(TGLBaseSceneObject(GeomsList[g1].Data),GeomsList[g1]);

  dJointGroupEmpty(ContactGroup1);
end;


procedure TForm1.CreateODE1;
begin
  ContactNum1   := 32;
  InitODE('');

  World1        := dWorldCreate();
  Space1        := dHashSpaceCreate(nil);
  dSpaceSetCleanup(Space1, 0);

  ContactGroup1 := dJointGroupCreate(0);
  dWorldSetGravity(World1, 0, 0, -0.55);
  dWorldSetCFM(World1, 1e-5);
end;


procedure TForm1.DestroyODE1;
begin
  dJointGroupDestroy(ContactGroup1);
  ContactGroup1 := 0;
  dSpaceDestroy(Space1);
  Space1 := nil;
  dWorldDestroy(World1);
  World1 := nil;
  dCloseODE;
end;


procedure nearCallback(Data : pointer; o1, o2 : PdxGeom); cdecl;
const
  cCOL_MAX = 32;
var
  i, NumC : Integer;
  b1, b2  : PdxBody;
  Contact : array [0..cCOL_MAX-1] of TdContact;
  C       : TdJointID;
begin
    // exit without doing anything if the two bodies are connected by a joint
  b1 := dGeomGetBody(o1);
  b2 := dGeomGetBody(o2);
  if     Assigned(b1)
     and Assigned(b2)
     and (dAreConnected(b1, b2) <> 0)
  then exit;

  for i := 0 to cCOL_MAX -1 do begin
    Contact[i].surface.mode       := ord(dContactBounce);
      // This determines friction, play around with it!
    Contact[i].surface.mu         := 0.9; //dInfinity; SHOULD BE INFINITY!
    Contact[i].surface.mu2        := 0;
    Contact[i].surface.bounce     := 0.8;
    Contact[i].surface.bounce_vel := 0.01;
  end;

  NumC := dCollide(o1, o2, cCOL_MAX,contact[0].geom, SizeOf(TdContact));
  for i := 0 to NumC -1 do begin
    C := dJointCreateContact(World1, ContactGroup1, @Contact[i]);
    dJointAttach(C, b1, b2);
  end;
end;







  // ----------------------------- service ------------------------------------

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  Caption := Format('%.1f FPS', [Viewer.FramesPerSecond]);
  Viewer.ResetPerformanceMonitor;
end;


procedure TForm1.ViewerMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if Shift = [ssLeft] then GLCamera1.MoveAroundTarget(mdy -y, mdx -x);
  mdx := x;
  mdy := y;
end;


procedure TForm1.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  if Viewer.Focused then
    GLCamera1.AdjustDistanceToTarget(Power(1.03, WheelDelta/120));
end;


procedure TForm1.FormResize(Sender: TObject);
begin
  GLCamera1.FocalLength := Width /10;
end;


procedure TForm1.ViewerMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Viewer.SetFocus;
end;

end.
